/**
 * @file baidu node bigpipe
 * @author liulangyu(liulangyu@baidu.com)
 * @date 2017-09-11
 */

// @see http://bp.baidu.com/doc/restapi.html

const axios = require('axios');
const qs = require('querystring');
const EventEmitter = require('events');
const util = require('./util');
const promiseRetry = require('promise-retry');

// 重试请求次数
const RETRY_TIMES = 1;

EventEmitter.defaultMaxListeners = 0;
axios.defaults.headers.post['Content-Type'] = 'application/x-www-form-urlencoded';

/**
 * rest api base path
 *
 * @const
 * @type {string}
 */
const REST_PATH = '/rest/pipe/';

const VALID_METHODS = ['publish', 'fetch', 'get', 'version'];

/**
 * 请求返回状态
 *
 * @const
 * @type {Object}
 */
const STATUS = {
    OK: 0,
    PIPE_NOT_EXIST: 55202,
    INVALID_PIPELET_ID: 59013,
    MSG_DUPLICATE: 110001,
    NO_USER: 55216,
    PASS_ERR: 20018,
    PERMISSION_DENIED: 55217,
    SIGN_TIME_EXPIRED: 110004
};

class Bigpipe extends EventEmitter {

    /**
     * constructor
     *
     * @constructor
     * @param {Object} conf bigpipe配置
     * @param {string} conf.url 请求地址 (查询：http://bp.baidu.com/account/webservice)
     * @param {string} conf.pipename 管道名称
     * @param {string} conf.username acl用户名
     * @param {string} conf.token webservice token
     */
    constructor(conf = {}) {
        super();

        // 参数校验
        for (let key of ['url', 'pipename', 'username', 'token']) {
            if (!conf.hasOwnProperty(key)) {
                throw new Error(`need args: ${key}`);
            }
        }

        this.conf = conf;
    }

    /**
     * normalize protocol
     *
     * @public
     * @param {string} url url
     * @return {string}
     */
    normalizeProtocol(url) {
        if (!/^https?:\/\//i.test(url)) {
            url = 'http://' + url;
        }

        return url;
    }

    /**
     * bigpipe request url
     *
     * @public
     * @param {(string | Object)=} query query
     * @return {string}
     */
    buildUrl(query = '') {
        const conf = this.conf;
        let url = this.normalizeProtocol(conf.url);

        url = `${url}${REST_PATH}${conf.pipename}`;

        util.isObject(query) && (query = qs.stringify(query));
        query && (url = `${url}?${query}`);

        return url;
    }

    /**
     * 默认请求传递的参数
     *
     * @public
     * @return {Object}
     */
    getDefaultReqData() {
        const conf = this.conf;
        // 过期时间设置为10分钟
        const expires = Math.round(Date.now() / 1000) + 600;

        return {
            expires,
            username: this.conf.username,
            sign: util.md5(expires + conf.token)
        };
    }

    /**
     * 请求方法是否有效
     *
     * @public
     * @param {string} method 请求方法
     * @return {boolean}
     */
    isMethodsValid(method) {
        return method && ~VALID_METHODS.indexOf(method);
    }

    /**
     * 发布消息
     *
     * @public
     * @param {number} pipeletid pipelet id
     * @param {Object} data 数据
     * @param {string} data.guid 消息的id，相同guid的消息将被视为重复的消息
     * @param {string} data.msg 消息内容
     * @return {Promise}
     */
    publish(pipeletid, data) {
        return promiseRetry(
            (retry, number) => new Promise((resolve, reject) => {
                axios
                    .post(
                        this.buildUrl({pipeletid}),
                        qs.stringify(
                            Object.assign(this.getDefaultReqData(), {
                                method: 'publish',
                                /* eslint-disable fecs-camelcase */
                                msg_guid: data.guid,
                                /* eslint-enable fecs-camelcase */
                                message: data.msg
                            })
                        )
                    )
                    .then(res => {
                        const {status, errmsg} = res.data;

                        if (status === STATUS.OK) {
                            resolve();
                        }
                        else {
                            reject(status, errmsg);
                        }
                    })
                    .catch(reject);
            }).catch(retry),
            {
                retries: RETRY_TIMES
            }
        );
    }

    /**
     * 主动获取消息
     *
     * @public
     * @param {number} pipelet pipelet id
     * @param {Object} data 数据
     * @param {string} data.id 消息的id，消息的id, -1表示获取最新一条消息，-2表示获取最旧的一条消息
     * @param {number=} data.batch 指定本次请求最多fetch的消息条数，默认为1，最大32
     * @return {Promise}
     */
    fetch(pipelet, data) {
        return new Promise((resolve, reject) => {
            axios
                .post(
                    this.buildUrl(),
                    qs.stringify(Object.assign(this.getDefaultReqData(), {method: 'fetch', pipelet, ...data}))
                )
                .then(res => {
                    // @param {Array.<Object>} messages
                    // 每条消息格式如下：
                    // {
                    //     pipe: 'string: 消息对应的pipename',
                    //     pipelet: 'number: 消息对应的pipeletid',
                    //     id: 'number: 消息的id',
                    //     data: 'string: 消息的数据'
                    // }
                    const {status, errmsg, messages} = res.data;

                    if (status === STATUS.OK) {
                        resolve(messages || []);
                    }
                    else {
                        reject(status, errmsg);
                    }
                })
                .catch(reject);
        });
    }

    /**
     * 查询pipe信息
     *
     * @public
     * @param {string} query 查询的内容，默认为pipeletnum（现在只支持这个）
     * @return {Promise}
     */
    get(query = 'pipeletnum') {
        return new Promise((resolve, reject) => {
            axios
                .post(
                    this.buildUrl(),
                    qs.stringify(Object.assign(this.getDefaultReqData(), {method: 'get', query}))
                )
                .then(res => {
                    const {status, errmsg} = res.data;

                    if (status === STATUS.OK) {
                        resolve(res.data[query]);
                    }
                    else {
                        reject(status, errmsg);
                    }
                })
                .catch(reject);
        });
    }

    /**
     * 查询版本号
     *
     * @public
     * @return {Promise}
     */
    version() {
        const url = this.normalizeProtocol(this.conf.url) + '/rest/version';

        return new Promise((resolve, reject) => {
            axios
                .get(url)
                .then(res => resolve(res.data.version))
                .catch(reject);
        });
    }
}

module.exports = Bigpipe;
